/* Go started 4/17/88 by Todd R. Johnson */
/* 8/8/89 cleaned up for first release */
/* Public Domain */

#include <cstdlib>
#include <iostream>
#include <iomanip>
#include <string>

using namespace std;

#include "goplayutils.h"
#include "amigo.h"

extern char   *playReason;
extern short  playLevel, showTrees;
extern char   *message;

game thegame; // the game class instantiated

/* List of vertices where handicap stones have been placed */
string handicapList;

struct bRec	goboard[maxSIZE][maxSIZE];	/*-- The main go board		--*/

struct Group	GroupList[MAXGROUPS];	/*-- The list of Groups		--*/
short		DeletedGroups[4];	/*-- Codes of deleted groups	--*/

short		GroupCount = 0;		/*-- The total number of groups	--*/
short		DeletedGroupCount;	/*-- The total number of groups	--*/
					/*-- deleted on a move		--*/
short		ko, koX, koY;
short		blackTerritory,whiteTerritory;
short		blackPrisoners, whitePrisoners;
short		showMoveReason	= FALSE,
		groupInfo	= FALSE,
		whitePassed	= FALSE,
		blackPassed	= FALSE;


/* Arrays for use when checking around a point */
short		xVec[4] = {0, 1, 0, -1};
short		yVec[4] = {-1, 0, 1, 0};

// two dummy functions

void placestone(enum bVal color, int x, int y) {
cout << "Placing stone, color: " << color << " column: " << x;
cout << " row: " << y << endl;
}

void removestone(int x, int y) {
cout << "Removing stone, column: " << x;
cout << " row: " << y << endl;
}

enum bVal nextp (enum bVal p) {

/*Return the code for the other player, who is the next to play.*/

enum bVal np = BLACK;

if (p == BLACK) np = WHITE;
else np = BLACK;

return np;
}

short int member(short group, short *grouplist, short cnt)
{
	unsigned short	i;

	for (i = 0; i < cnt; i++)
		if (grouplist[i] == group)
			return TRUE;
	return FALSE;
}

/* Does a stone at x, y connect to any groups of color? */
short int Connect( enum bVal color, short x, short y, short fGroups[], short *fCnt, short eGroups[], short *eCnt)
{
	unsigned short	point = 0;
	short		tx, ty, total = 0;
	enum bVal	opcolor = WHITE;


	*fCnt = 0;
	*eCnt = 0;
	if (color == WHITE)
		opcolor = BLACK;
	for (point = 0; point <= 3; point++ )  
	{
		tx = x + xVec[point];
		ty = y + yVec[point];
		if (!LegalPoint(tx,ty))
			continue;
		if (goboard[tx][ty].Val == color)
		{
			total++;
			if (!member(goboard[tx][ty].GroupNum, fGroups, *fCnt))
				fGroups[(*fCnt)++] = goboard[tx][ty].GroupNum;
		}
		else if (goboard[tx][ty].Val == opcolor)
		{
			total++;
			if (!member(goboard[tx][ty].GroupNum, eGroups, *eCnt))
				eGroups[(*eCnt)++] = goboard[tx][ty].GroupNum;
		}
	}
	return total;
}

/* Returns the maximum number of liberties for a given intersection */
short int Maxlibs(short x, short y)
{
	short	cnt = 4;


	if (x == 0 || x == (thegame.boardSize - 1))
		cnt--;
	if (y == 0 || y == (thegame.boardSize - 1))
		cnt--;
	return cnt;
}

void DeleteGroupFromStone( short x, short y )
{
	if (goboard[x][y].Val != EMPTY)
		GroupCapture(goboard[x][y].GroupNum);
}

/* Determine whether x, y is suicide for color */
short int Suicide(enum bVal color, short x, short y)
{
	enum bVal	opcolor = BLACK;
	short		friendlycnt, friendlygroups[4],
			enemycnt, enemygroups[4],
			total;
	short		maxlibs, i, libcnt = 0;


	if (color == BLACK)
		opcolor = WHITE;
	maxlibs = Maxlibs( x, y);
	total = Connect(color, x, y, friendlygroups, &friendlycnt,
					enemygroups, &enemycnt);

	if (total < maxlibs)
		return FALSE;

	/* Check for a capture */
	for (i = 0; i < enemycnt; i++)
		if (GroupList[enemygroups[i]].liberties == 1)
			return FALSE;
	for (i = 0; i < friendlycnt; i++)
		libcnt += (GroupList[friendlygroups[i]].liberties - 1);
	if (libcnt != 0)
		return FALSE;
	return TRUE;
}
   
/* Returns the number of liberties for x, y */
short int StoneLibs( short x, short y )
{
	short		cnt = 0, tx, ty;
	unsigned short	point;


	for (point = 0; point <= 3; point++)
	{
		tx = x + xVec[point];
		ty = y + yVec[point];
		if (LegalPoint(tx,ty) && goboard[tx][ty].Val == EMPTY)
			cnt++;
	}
	return cnt;
}
   
void
EraseMarks()
{
	short int x, y;

	for (x=0; x < thegame.boardSize; x++)
	  for (y=0; y < thegame.boardSize; y++)
			goboard[x][y].marked = FALSE;
}

/* Place a stone of color at x, y */
short int GoPlaceStone(enum bVal color, short x, short y)
{
	short	fgroups[4], egroups[4];	/* group codes surrounding stone */
	short	fcnt, ecnt, i;
	short	lowest = GroupCount + 1;


	DeletedGroupCount = 0;
	if (goboard[x][y].Val != EMPTY || Suicide(color,x,y))
		return FALSE;

	if (ko && koX == x && koY == y)
		return FALSE;

	ko = FALSE;
// placestone( color, x, y );
	goboard[x][y].Val = color;
	/* Does the new stone connect to any friendly stone(s)? */
	Connect(color, x, y, fgroups, &fcnt, egroups, &ecnt);
	if (fcnt)
	{
		/* Find the connecting friendly group with the lowest code */
		for (i = 0; i < fcnt; i++)
			if (fgroups[i] <= lowest)
				lowest = fgroups[i];
		/*-- Renumber resulting group --*/
		/*-- Raise the stone count of the lowest by one to account --*/
		/*-- for new stone --*/
		goboard[x][y].GroupNum = lowest;
		GroupList[lowest].count++;
		for (i = 0; i < fcnt; i++)
			if (fgroups[i] != lowest)
				MergeGroups(lowest, fgroups[i]);
		/* Fix the liberties of the resulting group */
		CountLiberties(lowest);
	}
	else
	{
		/* Isolated stone.  Create new group. */
		GroupCount++;
		lowest = GroupCount;
		GroupList[lowest].color = color;
		GroupList[lowest].count = 1;
		GroupList[lowest].internal = 0;
		GroupList[lowest].external = StoneLibs( x, y);
		GroupList[lowest].liberties = GroupList[lowest].external;
		GroupList[lowest].eyes = 0;
		GroupList[lowest].alive = 0;
		GroupList[lowest].territory = 0;
		goboard[x][y].GroupNum = lowest;
	}
	/* Now fix the liberties of enemy groups adjacent to played stone */
	FixLibs(color, x, y, PLACED);  /* Fix the liberties of opcolor */
	ReEvalGroups(color, x, y, lowest);
	RelabelGroups();
	return TRUE;
}

/* Remove a stone from the board */
void GoRemoveStone( short x, short y )
{
	goboard[x][y].Val = EMPTY;
	goboard[x][y].GroupNum = 0;
// removestone( x, y );
}

/* Merges two groups -- Renumbers stones and deletes second group from
list.  Fixes stone count of groups.  This does not fix anything else. 
FixLibs must be called to fix liberties, etc. */
void MergeGroups( short g1, short g2 )
{
	short	x, y;

	ForeachPoint(y,x)
		if (goboard[x][y].GroupNum == g2)
			goboard[x][y].GroupNum = g1;
	GroupList[g1].count += GroupList[g2].count;
	DeleteGroup( g2 );  /* Removes group from GroupList */
}

/* Stores a group code to be deleted */
void DeleteGroup(short code)
{
	DeletedGroups[DeletedGroupCount++] = code;
}

/* Re-evaluate the groups given the last move.  This assumes that the
last move has been merged into adjoining groups and all liberty counts
are correct.  Handles capture. Checks for Ko.  Keeps track of captured
stones. code is the group number of the stone just played. */
void ReEvalGroups(enum bVal color, short x, short y, short code)
{
	short		fgroups[4], egroups[4],
			fcnt, ecnt, i, killcnt = 0, count = 0;
	enum bVal	opcolor = BLACK;

	if (color == BLACK)
		opcolor = WHITE;
	/* Check for capture */
	Connect( color, x, y, fgroups, &fcnt, egroups, &ecnt);
	if (ecnt)
	{
		/* See if any of the groups have no liberties */
		for (i = 0; i < ecnt; i++)
			if (GroupList[egroups[i]].liberties == 0)
			{
				killcnt++;
				count = GroupList[egroups[i]].count;
				GroupCapture( egroups[i]);
			}
	}
	/* Check for ko.  koX and koY are set in GroupCapture above. */
	if (killcnt == 1 && count == 1 && GroupList[ code  ].count == 1
				&& GroupList[ code ].liberties == 1)
	{
		ko = TRUE;
	}
	if (killcnt)
//		intrPrisonerReport( blackPrisoners, whitePrisoners);
	/* Set eye count for groups */
	CountEyes();
}

/* Remove a captured group from the board and fix the liberties of any
   adjacent groups.  Fixes prisoner count. Sets KoX and KoY */
/*-- update display of captured stones -neilb --*/
void GroupCapture(short code)
{
	short	x, y;

	if (GroupList[code].color == BLACK)
		blackPrisoners += GroupList[code].count;
	else
		whitePrisoners += GroupList[code].count;
//	intrPrisonerReport(blackPrisoners, whitePrisoners);
	ForeachPoint(y,x)
		if (goboard[x][y].GroupNum == code)
		{
			FixLibs(GroupList[code].color,x,y,REMOVED);
			GoRemoveStone(x, y);
			koX = x;
			koY = y;
		}
	DeleteGroup( code);
}

/* Fix the liberties of groups adjacent to x, y.  move indicates
  whether a stone of color was placed or removed at x, y
  This does not change liberty counts of friendly groups when a stone
  is placed.  Does not do captures. */
void FixLibs( enum bVal color, short x, short y, short move)
{
	short	fgroups[4], fcnt, egroups[4], ecnt, i;
	enum	bVal opcolor = BLACK;

	if (color == BLACK)
		opcolor = WHITE;
	Connect( color, x, y, fgroups, &fcnt, egroups, &ecnt);
	if (move == PLACED)
		for (i = 0; i < ecnt; i++)
			GroupList[egroups[i]].liberties--;
	else /* Stone removed so increment opcolor */
		for (i = 0; i < ecnt; i++)
			GroupList[egroups[i]].liberties++;
}

void place_handicap_stone (int x, int y) {
  GoPlaceStone( BLACK, x, y );
  // add vertex to handicapList
  handicapList += (char)lettercol(x);
  if ( y >= 9 ){
    handicapList += "1";
    if ( y == 9 )
      handicapList += "0";
    else
      handicapList += ('1' + ( y % 10 ));
  }
  else
    handicapList += ( '1' + y );
  handicapList += " ";
}

int colletter(int letter)
/*
Return the column # (0 to boardsize-1) which corresponds to letter.
Due to perversity of tournament organizers etc., 'i' or 'I' is
omitted, leading to an increase in complexity of this routine.
Return (-1) on illegal input.
*/
{ register int result;

  if('a'<=letter&&letter<='z')
    result= letter-'a';
  else if('A'<=letter&&letter<='Z')
    result= letter-'A';
  else
    return (-1);

  if(8>result)
    ;
  else if(8==result)
    return (-1);
  else
    --result;

  if(result >= thegame.boardSize)
    return (-1);
  else
    return result;
}

void initgame() {

  // Initialize the Random Number Generator
  srand(thegame.RANDOM_SEED);

  // Initialize various thegame variables
  // ko information
  thegame.kox = thegame.koy = (-1);
  // Captured pieces
  thegame.capturedWhite = thegame.capturedBlack = 0;
  // turn
  thegame.tur = 1;
  // pass status
  thegame.qpa = 0;
  // who is to play
  if(thegame.blackHandicap < 2)
    thegame.pla= BLACK;
  else
    thegame.pla= WHITE;

  // Amigo initialization  
  goRestart();
}  

void goRestart()
{
	short int x, y;

	GroupCount = 0;
	ko = FALSE;
	blackPrisoners = whitePrisoners = 0;
  // Erase the board
	for (x = 0; x < thegame.boardSize ; x++)
    for (y = 0; y < thegame.boardSize ; y++) {
      goboard[x][y].Val = EMPTY;
      goboard[x][y].GroupNum = 0;
    }

  place_fixed_handicap();
}


void place_fixed_handicap()
{
  // place the stones based on board size and build
  // the handicap stones vertices list simultaneously

  // clear the handicap stones vertices list first
  handicapList.erase();

  if(thegame.blackHandicap > 1){
    if (thegame.boardSize == 9)
      setHandicap9(thegame.blackHandicap);
    if (thegame.boardSize == 13)
      setHandicap13(thegame.blackHandicap);
    if (thegame.boardSize == 19)
      setHandicap19(thegame.blackHandicap);
  }
}

void setHandicap9(int handicap9) {
 if (handicap9 < 2)
    return;

  place_handicap_stone(2, 2);
  place_handicap_stone(6, 6);

  if (handicap9 >= 3)
    place_handicap_stone(2, 6);
  if (handicap9 >= 4)
    place_handicap_stone(6, 2);
}

void setHandicap13(int handicap13) {
 if (handicap13 < 2)
    return;

  place_handicap_stone(3, 3);
  place_handicap_stone(9, 9);

  if (handicap13 >= 3)
    place_handicap_stone(3, 9);
  if (handicap13 >= 4)
    place_handicap_stone(9, 3);
  if (handicap13 == 5 || handicap13 == 7 || handicap13 == 9)
    place_handicap_stone(6, 6);
  if (handicap13 >= 6) {
    place_handicap_stone(9, 6);
    place_handicap_stone(3, 6);
  }
  if (handicap13 >= 8) {
    place_handicap_stone(6, 9);
    place_handicap_stone(6, 3);
  }
}

void setHandicap19(int handicap19) {
 if (handicap19 < 2)
    return;

  place_handicap_stone(3, 3);     // D4
  place_handicap_stone(15, 15);   // Q16

  if (handicap19 >= 3)
    place_handicap_stone(3, 15);  // D16
  if (handicap19 >= 4)
    place_handicap_stone(15, 3);  // Q4
  if (handicap19 == 5 || handicap19 == 7 || handicap19 == 9)
    place_handicap_stone(9, 9);   // K10
  if (handicap19 >= 6) {
    place_handicap_stone(3, 9);   // D10
    place_handicap_stone(15, 9);  // Q10
  }
  if (handicap19 >= 8) {
    place_handicap_stone(9, 3);   // K4
    place_handicap_stone(9, 15);  // K16
  }
}

void opponentpass()
{
  if (thegame.amigoColor == BLACK) whitePassed = TRUE;
  else blackPassed = TRUE;
  thegame.qpa= 1;
  // clear the ko status
  ko = FALSE;
  // done
  movedone();  
}

void movedone()
/*
  Do everything which must be done to complete a move after the stone
  is placed.
 */
{ thegame.pla= nextp(thegame.pla);
  ++thegame.tur;
}

int mymove()
{
  int move;
  move = enginemove();
  if ( (move == PASS) || (move == BOTHPASS) ){
    cout << ((thegame.amigoColor == WHITE)?"White":"Black");
    cout << " passes." << endl;
  }
  else {
    cout << ((thegame.amigoColor == WHITE)?"White":"Black");
    cout << " moves to ";
    cout.put(lettercol(thegame.ex));
    cout << thegame.ey+1 << "." << endl;
  }
  return move;
}

int enginemove()
/*Calculate and execute the engine's move.*/
{
  short int x, y;   // coords of the move selected

  if (genMove(thegame.amigoColor, &x, &y)) {
    // if amigo found a move, play it
    GoPlaceStone(thegame.amigoColor, x, y);
    // update pass information
    if (thegame.amigoColor == BLACK) blackPassed = FALSE;
    else whitePassed = FALSE;
    // update game object
    thegame.qpa= 0;
    thegame.ex = x;
    thegame.ey = y;
    // write vertex to string
    thegame.emove = (char)lettercol(x);
    if ( y >= 9 ){
      if ( y >= 19 )
        thegame.emove += "2";
      else
        thegame.emove += "1";
      if ( y == 9 || y == 19 )
        thegame.emove += "0";
      else
        thegame.emove += ('1' + ( y % 10 ));
    }
    else
      thegame.emove += ( '1' + y );
    return PLAY_MOVE;
  }
  else {
    if (thegame.amigoColor == BLACK) blackPassed = TRUE;
    else whitePassed = TRUE;
    thegame.emove = "pass";
    if(thegame.qpa) return BOTHPASS;
    else {
      thegame.qpa= 1;
      // clear the ko status?
      return PASS;
    }
  }
  movedone(); // the engine is done
}



/* if any groups have been deleted as a result of the last move, this
   routine will delete the old group numbers from GroupList and
   reassign group numbers. */
void RelabelGroups()
{
	unsigned	short i, j, x, y;

	for (i = 0; i < DeletedGroupCount; i++)
	{
		/* Relabel all higher groups */
		ForeachPoint(y,x)
			if (goboard[x][y].GroupNum > DeletedGroups[i])
				goboard[x][y].GroupNum--;
		/* Move the groups down */
		for (y = DeletedGroups[i]; y < GroupCount; y++)
			GroupList[y] = GroupList[y+1];
		/* fix the group numbers stored in the deleted list */
		for (j = i+1; j < DeletedGroupCount; j++)
			if (DeletedGroups[j] > DeletedGroups[i])
				DeletedGroups[j]--;
		GroupCount--;
	}
}

/* Returns liberty count for x, y intersection.  Sets marked to true
   for each liberty */
short int CountAndMarkLibs( short x, short y )
{
	short	tx,ty,i;
	short	cnt = 0;


	for (i=0;i<4;i++)
	{
                tx = x + xVec[i];
                ty = y + yVec[i];
                if (LegalPoint(tx,ty) && goboard[tx][ty].Val == EMPTY
					&& goboard[tx][ty].marked == FALSE)
		{
			cnt++;
			goboard[tx][ty].marked = TRUE;
		}
	}
	return cnt;
}

/* Determine the number of liberties for a group given the group code
   num */
void CountLiberties( short code )
{
	short	x, y, libcnt = 0;

	ForeachPoint(y,x)
		if (goboard[x][y].GroupNum == code)
			libcnt += CountAndMarkLibs( x, y);
	EraseMarks();
	GroupList[code].liberties = libcnt;
}

void CheckForEye( short x, short y, short groups[], short cnt, short *recheck)
{
	short	i;

	for (i = 0; i < (cnt-1); i++)
		if (groups[i] != groups[i+1])
		{
			/* Mark liberty for false eye check */
			goboard[x][y].marked = TRUE;
			(*recheck)++;
			return;
		}
	/* It is an eye */
	GroupList[groups[i]].eyes += 1;
}
   
/* Set the eye count for the groups */
void CountEyes()
{
	short	i, x, y,
		wgroups[4], bgroups[4], wcnt, bcnt, max, cnt, recheck = 0, eye;

	for (i = 1; i <= GroupCount; i++)
		GroupList[i].eyes = 0;

	ForeachPoint(y,x)
	{
		if (goboard[x][y].Val != EMPTY)
			continue;
		cnt = Connect(WHITE,x,y,wgroups,&wcnt,bgroups,&bcnt);
		max = Maxlibs( x, y);
		if (cnt == max && wcnt == 1 && bcnt == 0)
			GroupList[wgroups[0]].eyes++;
		else if (cnt == max && bcnt == 1 && wcnt == 0)
			GroupList[bgroups[0]].eyes++;
		else if (cnt == max && ( bcnt == 0 || wcnt == 0 ))
		{
			goboard[x][y].marked = TRUE;
			recheck++;
		}
	}

	/*-- Now recheck marked liberties to see if two or more one eye --*/
	/*-- groups contribute to a false eye */
	if (recheck == 0)
		return;

	ForeachPoint(y,x)
		if (goboard[x][y].marked)
		{
			recheck--;
			goboard[x][y].marked = FALSE;
			Connect( WHITE, x, y, wgroups, &wcnt, bgroups, &bcnt);
			/* If all the groups have at least one eye then all the
			groups are safe from capture because of the common
			liberty at x, y */
			eye = TRUE;
			for (i = 0; i < wcnt; i++)
				if (GroupList[wgroups[i]].eyes == 0)
					eye = FALSE;
			if (eye)
				for (i = 0; i < wcnt; i++)
					GroupList[wgroups[i]].eyes++;
			for (i = 0; i < bcnt; i++)
				if (GroupList[bgroups[i]].eyes == 0)
					eye = FALSE;
			if (eye)
				for (i = 0; i < bcnt; i++)
					GroupList[bgroups[i]].eyes++;
			if (recheck == 0)
				return;
		}
}


short	foo[maxSIZE][maxSIZE];

// Scoring

void score()
{
  // Count territories
  estimatescore();
  
  // Output results
  cout_score();
}


void cout_score()
{
  float finalScore;

  cout << "(the end of the game)\n";
  cout << "Komi is: " << setprecision(1) << thegame.komi << endl;
  cout << "Points count is:\n";
  // Black
  cout << "Black has " << thegame.territoryBlack;
  cout << " territory, and has captured " << thegame.capturedWhite;
  cout << " White stones." << endl;
  // White
  cout << "White has " << thegame.territoryWhite;
  cout << " territory, and has captured " << thegame.capturedBlack;
  cout << " Black stones." << endl;
  // Who wins?
  finalScore = thegame.territoryBlack + thegame.capturedWhite
             - thegame.territoryWhite - thegame.capturedBlack
             - thegame.komi;
  if( 0 == finalScore )
    cout << "(a tie)\n"; // impossible
  else if( finalScore > 0 )
    cout << "(Black wins.)\n";
  else
    cout << "(White wins.)\n";
}


void estimatescore()
{
	short	x,y;
	int	btotal, wtotal;

	blackTerritory = whiteTerritory = 0;
	ForeachPoint(y,x)
	{
		goboard[x][y].marked = FALSE;
		foo[x][y] = CNT_UNDECIDED;
	}
	ForeachPoint(y,x)
		if (goboard[x][y].Val==EMPTY && foo[x][y]==CNT_UNDECIDED)
		{
			FillPoints(x,y,CountFromPoint(x,y));
		}

	wtotal = whiteTerritory + blackPrisoners;
	btotal = blackTerritory + whitePrisoners;

  thegame.territoryBlack = blackTerritory;
  thegame.territoryWhite = whiteTerritory;

  thegame.capturedBlack = blackPrisoners;
  thegame.capturedWhite = whitePrisoners;
}

void FillPoints(short x, short y, short val)
{
	int	i;
	short	tx,ty;

	if ((foo[x][y] = val) == CNT_BLACK_TERR)
		blackTerritory++;
	else if (val == CNT_WHITE_TERR)
		whiteTerritory++;
	for (i=0;i<4;i++)
	{
		tx = x + xVec[i];
		ty = y + yVec[i];
		if (!LegalPoint(tx,ty))
			continue;
		if (goboard[tx][ty].Val==EMPTY && foo[tx][ty]==CNT_UNDECIDED)
			FillPoints(tx,ty,val);
	}
}

short int CountFromPoint( short x, short y )
{
	int	i;
	short	tx,ty;
	short	blkcnt=0,whtcnt=0;
	short	baz;

	goboard[x][y].marked = TRUE;
	for (i=0;i<4;i++)
	{
		tx = x + xVec[i];
		ty = y + yVec[i];
		if (!LegalPoint(tx,ty))
			continue;
		if (goboard[tx][ty].Val == BLACK)
			blkcnt++;
		else if (goboard[tx][ty].Val == WHITE)
			whtcnt++;
		else
		{
			if (goboard[tx][ty].marked)
				continue;
			baz = CountFromPoint(tx,ty);
			if (baz == CNT_NOONE)
				return CNT_NOONE;
			else if (baz == CNT_BLACK_TERR)
				blkcnt++;
			else if (baz == CNT_WHITE_TERR)
				whtcnt++;
		}
		if (blkcnt && whtcnt)
			return CNT_NOONE;
	}
	if (blkcnt && !whtcnt)
		return CNT_BLACK_TERR;
	else if (whtcnt && !blkcnt)
		return CNT_WHITE_TERR;
	else
		return CNT_UNDECIDED;
}

void printgame()
/*
Output the game board theboard (with ko info from thegame) to the console.
Currently the style is:

   a b c d e f g h j
 1 . . . . . . . . . 1
 2 . . . . . O O . . 2
 3 . . O . . . # . . 3
 4 . . O . . . # . . 4
 5 O O ^ O . . . . . 5
 6 . # O # . # # . . 6
 7 . # # # . . . . . 7
 8 . . . . . . . . . 8
 9 . . . . . . . . . 9
   a b c d e f g h j

where '^' is a ko, '#' is black, 'O' is white.
*/
{ int x, y;
  // update thegame ko information
  if (ko) {
    thegame.kox = koX;
    thegame.koy = koY;
  }
  else thegame.kox = thegame.koy = (-1);
   
  // update thegame captured pieces information
  thegame.capturedBlack = blackPrisoners;
  thegame.capturedWhite = whitePrisoners;
  
  // print the board
  cout << "  "; // avoid printing an empty line for GTP V2
  cout << endl << "  ";
  for(x=0; x<thegame.boardSize; ++x) {
    cout << " ";
    cout.put(lettercol(x));
  }
  cout << "\n";
  for(y=0; y<thegame.boardSize; ++y)
  { cout.setf( ios::right, ios::adjustfield );
    cout.width(2);
    cout << (1+y) << " ";
    for(x=0; x<thegame.boardSize; ++x) {
      enum bVal here = goboard[x][y].Val;
      cout << (( thegame.kox == x && thegame.koy == y ) ? '^' :
      (EMPTY==here) ? '.' :
      (WHITE==here) ? 'O' :
      (BLACK==here) ? '#' :
      // Replaced
      // panic("illegal board type in printboard()"));
      // by
      '@');  // call the panic function later
      if ( x != (thegame.boardSize-1) ) cout << " ";
    }
    cout << " " << flush;
    cout.unsetf( ios::right );
    cout.width(2);
    cout.setf( ios::left, ios::adjustfield );
    cout << (y+1);
    cout.unsetf( ios::left );
    if ( y == 0 )
      cout << "   BLACK (#) has captured " << thegame.capturedWhite << " stones.";
    if ( y == 1 )
      cout << "   WHITE (O) has captured " << thegame.capturedBlack << " stones.";
    cout << endl;
  }
  cout << flush;
  cout << "  ";
  for(x=0; x<thegame.boardSize; ++x) {
    cout << " ";
    cout.put(lettercol(x));
  }
  cout << endl;
}

char getnb()
/*Read and return the first nonblank character from the standard input.*/
{ char c;
  const char TAB = 8;
  do
    c = cin.get();
  while(' '==c||'\n'==c||TAB==c);
  return c;
}

int scanfcoo(int *x, int *y)
/*
Read a coordinate from the standard input and write it to *x and *y.  Like
scanf, return 1 for success, 0 for failure, negative numbers for special
moves, e.g. PASS and RESIGN.
On any input error, the function eats input up to the next newline.
*/
{
  int xx, yy;
  char c1, c2, c3;

/* Get first nonblank character of input, normally a letter code for a column. */
  c1 = getnb();

/* Check for possible resignation (a '.' at the start of the input). */
  if('.'==c1)
  {
    do c3 = cin.get();
    while( c3 != '\n' ); /* Skip to end of line. */
    return RESIGN;
  }

/* Check for possible "pass". */
  if('p'==lowercase(c1))
  { c2 = cin.get();
    if('a'!=lowercase(c2))
      cin.putback(c2);
    else {
      c2 = cin.get();
      if( 's' != lowercase(c2) ) {
        cin.putback(c2);
        goto skipquit;
      }
      else {
        c2 = cin.get();
        if( 's' != lowercase(c2) ) {
            cin.putback(c2);
            goto skipquit;
          }
        else
          return PASS;
      }
    }
  }
  xx = colletter(c1);
  if(!(cin >> yy))       // get a valid integer input, otherwise return 0
  {
skipquit:                /* Skip to the next newline and return failure.*/
    do c3 = cin.get();
    while( c3 != '\n' ); /* Skip to end of line.*/
    return 0;
  }
  if(0==LegalPoint(xx,yy-1)) /* The -1 converts between C arrays starting */
                          /* at 0 and human boards counting from 1.*/
    goto skipquit;
  else
  { *x= xx;
    *y= yy-1;
    return 1;
  }
}

int enemymove()
/*Read and execute opponent's move.*/
{ int x, y;	/*coordinates of move*/
  int s;	/*return value of scanfcoo()*/

endrecurse:

  printgame();
  cout << "Game turn= " << thegame.tur << ".  ";
  cout << ((thegame.pla == WHITE)?"White":"Black") << " to play." << endl;
  cout << "Please input your move (" << ((thegame.pla == WHITE)?"White":"Black");
  cout << " #" << thegame.tur << "): ";
  s= scanfcoo(&x, &y);	/*Try to get a move.*/
  if(RESIGN==s)		/*If opponent resigns*/
   return RESIGN;
  if(PASS==s) {	/*If opponent passes*/
    movedone();
    thegame.kox = thegame.koy = (-1);
    if(thegame.qpa) return BOTHPASS;
    else
    { thegame.qpa= 1;
      return PASS;
    }
  }
  if( 0==s )
  { cout << "Please try again.  Input your move as a letter for the column, \n";
    cout << " followed by a number for the row number.\n";
    goto endrecurse;
  }
  if( 0==LegalPoint(x,y) )
  { cout << "Only moves which are on the board are legal.  Please try again.\n";
    goto endrecurse;
  }
  if( EMPTY != goboard[x][y].Val )
  { cout << "You cannot stack a stone on another.  ";
pleasetryanothermove:
    cout << "Please try another move." << endl;
    goto endrecurse;
  }
  if((x == koX) && (y == koY))
  { cout << "That move would violate the rule of ko.  ";
    goto pleasetryanothermove;
  }
  GoPlaceStone(thegame.pla, (short)x, (short)y);
  thegame.qpa= 0;
    cout.put(lettercol(x));
    cout << y+1 << endl;

  movedone();
  return 0;
}


